/* Silly program to "trivially" (ha!) store uid/gid numbers in correct place.
 * Version 2: Version 1 didn't work with LinuxPPC Release 5.
 * Version 3: Modified (formatting) + two small tweaks to get it to work on
 *            Solaris from version provided by
 *            Pascal DeMilly <pascal.contribs@newgenesys.com>
 *            "Needed a major cleanup as I saw too many cases where it
 *             will just not work".
 * Version 4: Modified to work even if mrkb is after mrke
 *            (true in some linkers (like in RHEL4)).
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>

int filebrand( int file, char *filename, int mrkb, int mrke );
void brand( int file, off_t pos, int direction );
off_t brandit( int file, int direction, off_t pos, int data );
int checkEndMarker( int file, off_t pos, int data );
int setuser( int *addr, char *username );
int setgroup( int *addr, char *groupname );
void usage();

/* Make sure this matches with auto_uids.c */
int mrkb = 0x6d726b62; /* Should be "mrkb" */
int auto_uida = 1;
int auto_uidd = 2;
int auto_uidl = 3;
int auto_uido = 4;
int auto_uidp = 5;
int auto_uidq = 6;
int auto_uidr = 7;
int auto_uids = 8;
int auto_gidq = 9;
int auto_gidn = 10;
int mrke = 0x6d726b65; /* Should be "mrke" */


int main( int argc, char **argv )
{
  int n;
  int fd;

  /* Set the users and groups */
  setuser( &auto_uida, "alias"  );
  setuser( &auto_uidd, "qmaild" );
  setuser( &auto_uidl, "qmaill" );
  setuser( &auto_uido, "root"   );
  setuser( &auto_uidp, "qmailp" );
  setuser( &auto_uidq, "qmailq" );
  setuser( &auto_uidr, "qmailr" );
  setuser( &auto_uids, "qmails" );

  setgroup( &auto_gidq, "qmail"   );
  setgroup( &auto_gidn, "nofiles" );

  if ( argc < 2 )
  {
    usage();
  }

  for ( n = 1; n < argc; n++ )
  {
    if ( (fd = open( argv[n], O_RDWR )) == -1 )
    {
      fprintf( stderr, "Can't open file %s\n", argv[n] );
      return 1;
    }

    /* fprintf( stderr, "Searching %s\n", argv[n] ); */
    if ( filebrand( fd, argv[n], mrkb, mrke ) )
    {
      return 1;
    }

    close( fd );
  }

  return 0;
}


void usage( void )
{
  fprintf( stderr, "Usage: brand file ...\n" );
  exit( 1 );
}


/* Search for beginning and end marker and brand integers in between */
int filebrand( int file_, char *filename_, int mrkb_, int mrke_ )
{
  unsigned char buf [ sizeof(int) * 256 ]; // Buffer must be at least 1 int long.
  int n;
  int pos = 0;
  int marker;

  /* Go through each character of file, filling in buffer until string terminator
   found or buffer is full */
  while ( (n = read( file_, &buf[0], sizeof(buf) )) > sizeof(int) )
  {
    int i;

    /* Check buffer 1 char at a time looking for int that resembles mrkb.
       Stop before the end so we don't get a partial int.
    */
    for ( i = 0; i < (n - sizeof(int) + 1); i++, pos++ )
    {
      memcpy( &marker, &buf[i], sizeof(marker) );

      if ( marker == mrkb_ )
      {
        if ( checkEndMarker( file_, pos + 11 * sizeof(int), mrke_ ) )
        {
          brand( file_, pos, 1 );
          return 0;
        }
        else
        {
          pos++;
          break;
        }
      }
      else if ( marker == mrke_ )
      {
        // The linker linked it backwards
        if ( checkEndMarker( file_, pos + 11 * sizeof(int), mrkb_ ) )
        {
          brand( file_, pos, -1 );
          return 0;
        }
        else
        {
          pos++;
          break;
        }
      }
    }

    /* Now reposition; we could have read a part of a int at the end of the
       buffer.  Also, if we had found the marker the position would have
       changed to the marker end + 1.
    */
    lseek( file_, pos, SEEK_SET );
  }

  /* We should have found SOMETHING! */
  fprintf( stderr, "%s: Marker not found.\n", filename_ );

  return 1;
}


setuser( int *addr, char *username )
{
  struct passwd *myp;

  if ( (myp = getpwnam( username )) == NULL )
  {
    fprintf( stderr, "Can't find user %s\n", username );
    exit( 1 );
  }

  *addr = (int) (myp->pw_uid);
}


setgroup( int *addr, char *groupname )
{
  struct group *myg;

  if ( (myg = getgrnam( groupname )) == NULL )
  {
    fprintf( stderr, "Can't find group %s\n", groupname );
    exit( 1 );
  }

  *addr = (int) (myg->gr_gid);
}

void brand( int file_, off_t pos_, int direction_ )
{
  if ( direction_ == -1 )
  {
    pos_ += 11 * sizeof(int);
  }

  /* Make sure this matches with auto_uids.c */
  pos_ = brandit( file_, direction_, pos_, mrkb      ); /* Should be "mrkb" */
  pos_ = brandit( file_, direction_, pos_, auto_uida );
  pos_ = brandit( file_, direction_, pos_, auto_uidd );
  pos_ = brandit( file_, direction_, pos_, auto_uidl );
  pos_ = brandit( file_, direction_, pos_, auto_uido );
  pos_ = brandit( file_, direction_, pos_, auto_uidp );
  pos_ = brandit( file_, direction_, pos_, auto_uidq );
  pos_ = brandit( file_, direction_, pos_, auto_uidr );
  pos_ = brandit( file_, direction_, pos_, auto_uids );
  pos_ = brandit( file_, direction_, pos_, auto_gidq );
  pos_ = brandit( file_, direction_, pos_, auto_gidn );
  pos_ = brandit( file_, direction_, pos_, mrke      ); /* Should be "mrke" */
}


off_t brandit( int file_, int direction_, off_t pos_, int data_ )
{
  if ( direction_ == 0 )
  {
    fprintf( stderr, "brandit: Invalid direction\n" );
    exit( 1 );
  }

  if ( lseek( file_, pos_, SEEK_SET ) != pos_ )
  {
    perror( "Error: brandit(seek1)" );
    exit( 1 );
  }

  if ( write( file_, &data_, sizeof(data_) ) != sizeof(data_) )
  {
    perror( "Error: brandit(write)" );
    exit( 1 );
  }

  if ( direction_ > 0 )
  {
    pos_ += sizeof( int );
  }
  else
  {
    pos_ -= sizeof(int); // Because we've already moved forward 1 int.
  }
  
  return pos_;
}


int checkEndMarker( int file_, off_t pos_, int data_ )
{
  int compData;

  // Position to correct place.
  if ( lseek( file_, pos_, SEEK_SET ) != pos_ )
  {
    perror( "Error: checkEndMarker(seek1)" );
    exit( 1 );
  }

  // Read the data.
  if ( read( file_, &compData, sizeof(compData) ) != sizeof(compData) )
  {
    perror( "checkEndMarker:" );
    exit( 1 );
  }

  // Repositon back to original place.
  if ( lseek( file_, pos_, SEEK_SET ) != pos_ )
  {
    perror( "Error: checkEndMarker(seek1)" );
    exit( 1 );
  }

  // Return results.
  return data_ == compData;
}
